import click
from click_option_group import optgroup
from sarif_manager.sarif.azure_sarif import AzureSarif


@click.group("azure")
def azure():
    """Manage Azure DevOps artifacts."""


@azure.command("write-logs", help="Write discovered findings to logs in Azure Pipelines.")
@click.argument("sarif-file", type=click.Path(exists=True))
@optgroup.group("Repository Details")
@optgroup.option("--org", "organization", help="Repository URL")
@optgroup.option("--project", help="Azure DevOps project name")
def write_logs(sarif_file, organization, project):
    """Write discovered findings to logs in Azure Pipelines."""
    azure_sarif = AzureSarif()
    azure_sarif.set_metadata(organization, project)
    azure_sarif.load_from_file(sarif_file)
    azure_sarif.print_logging_commands()


@azure.command("create-work-items", help="Create work items in Azure DevOps for each finding.")
@click.argument("sarif-file", type=click.Path(exists=True))
@click.option("--write-logs", is_flag=True, help="Write discovered findings to logs in Azure Pipelines.")
@optgroup.group("Repository Details")
@optgroup.option("--org", "organization", help="Repository URL", required=True)
@optgroup.option("--project", help="Azure DevOps project name", required=True)
@optgroup.group("Azure DevOps Authentication")
@optgroup.option("--token", "personal_access_token", help="Azure DevOps personal access token", envvar="AZURE_DEVOPS_ACCESS_TOKEN", required=True)
def create_work_items(sarif_file, write_logs, organization, project, personal_access_token):
    """Create work items in Azure DevOps for each finding."""
    azure_sarif = AzureSarif()
    azure_sarif.set_metadata(organization, project)
    if write_logs:
        azure_sarif.print_logging_commands()
    azure_sarif.set_auth(personal_access_token)
    azure_sarif.load_from_file(sarif_file)
    azure_sarif.create_work_items()



import json
from typing import Union
from base64 import b64encode

from sarif_manager.sarif.azure_sarif_finding import AzureSarifFinding
from sarif_manager.sarif.azure_sarif_utils import create_work_item


class AzureSarif:
    def __init__(self):
        self.sarif_data = {}
        self.sarif_findings = []
        self.pat: Union[str, None] = None
        self.azure_headers: Union[dict, None] = None
        self.organization: Union[str, None] = None
        self.project: Union[str, None] = None
        self.repo_url: Union[str, None] = None

    def load_from_file(self, sarif_file: str):
        with open(sarif_file, 'r') as file:
            self.sarif_data = json.load(file)
        self.parse()

    def set_metadata(self, organization: str, project: str):
        self.organization = organization
        self.project = project
        self.repo_url = f'https://dev.azure.com/{organization}/_git/{project}'

    def set_auth(self, pat: str):
        self.pat = pat
        # Headers for Azure DevOps API
        credentials = b64encode(bytes(f":{pat}", 'utf-8')).decode('ascii')
        self.azure_headers = {
            'Authorization': f'Basic {credentials}',
            'Content-Type': 'application/json-patch+json'
        }

    @property
    def azure_api_url(self):
        return f"https://dev.azure.com/{self.organization}/{self.project}/_apis/wit/workitems/$Issue?api-version=6.0"

    def parse(self):
        for run in self.sarif_data.get('runs', []):
            for result in run.get('results', []):
                # fmt: off
                location = result.get('locations', [{}])[0].get('physicalLocation', {}).get('artifactLocation', {}).get('uri')
                line = result.get('locations', [{}])[0].get('physicalLocation', {}).get('region', {}).get('startLine', 0)
                description = "No description available."
                for rule in run['tool']['driver']['rules']:
                    if rule['id'] == result.get('ruleId'):
                        description = rule['fullDescription']['text']
                        break
                # fmt: on
                finding = {
                    'rule_id': result.get('ruleId'),
                    'message': result.get('message', {}).get('text'),
                    'severity': result.get('level'),
                    'location': location,
                    'line': line,
                    'description': description,
                }
                self.sarif_findings.append(finding)
        return self.sarif_findings

    def azure_devops_messages(self):
        """
        To log the findings in Azure DevOps pipelines as warnings or errors, you need to format the messages as
        Azure DevOps logging commands. This method generates the logging commands for each finding.

        More details: https://learn.microsoft.com/en-us/azure/devops/pipelines/scripts/logging-commands?view=azure-devops&tabs=bash
        """
        messages = set()
        for item in self.sarif_findings:
            item['repo_url'] = self.repo_url
            this_finding = AzureSarifFinding(**item)
            if this_finding.exclude:
                continue
            messages.add(this_finding.azure_devops_message)
        messages = list(messages)
        messages.sort()
        return messages

    def print_logging_commands(self):
        """Print the Azure DevOps logging commands to the console."""
        messages = self.azure_devops_messages()
        for message in messages:
            print(message)

    def create_work_items(self):
        """Create work items in Azure DevOps for each finding."""
        for item in self.sarif_findings:
            item['repo_url'] = self.repo_url
            this_finding = AzureSarifFinding(**item)
            if this_finding.exclude:
                continue
            create_work_item(
                rule_id=this_finding.rule_id,
                issue_title=this_finding.message,
                issue_description=this_finding.description,
                api_url=self.azure_api_url,
                headers=self.azure_headers
            )



import sys
import re
import markdown
import requests
from bs4 import BeautifulSoup


def trim_uuid(input_string: str) -> str:
    # Regex pattern to match a UUID (with or without hyphens)
    uuid_pattern = r'[a-f0-9]{8}-[a-f0-9]{4}-[a-f0-9]{4}-[a-f0-9]{4}-[a-f0-9]{12}'

    # Replace the first occurrence of the UUID and the following hyphen with an empty string
    trimmed_string = re.sub(uuid_pattern + r'-?', '', input_string, count=1)

    return trimmed_string.strip()


def create_work_item(
        rule_id: str, issue_title: str, issue_description: str, api_url: str, headers: dict
):
    """
    Create a single work item in Azure DevOps with URLs converted to HTML links.

    Reference: https://learn.microsoft.com/en-us/rest/api/azure/devops/?view=azure-devops-rest-7.2&viewFallbackFrom=azure-devops-rest-5.1

    """
    # Convert Markdown to HTML
    html_description = markdown.markdown(issue_description)

    # Convert URLs into clickable links
    url_pattern = re.compile(
        r'https?://[^\s<>"]+|www\.[^\s<>"]+'
    )
    html_description = url_pattern.sub(
        lambda x: f'<a href="{x.group(0)}">{x.group(0)}</a>', html_description
    )

    soup = BeautifulSoup(html_description, 'html.parser')
    formatted_description = soup.prettify()

    json_data = [
        {"op": "add", "path": "/fields/System.Title", "value": issue_title},
        {"op": "add", "path": "/fields/System.Description", "value": formatted_description},
        {"op": "add", "path": "/fields/System.Tags", "value": "Security Vulnerability"}
    ]
    response = requests.post(api_url, headers=headers, json=json_data)
    if response.status_code == 200:
        response_data = response.json()
        html_url = response_data['_links']['html']['href']
        print(f"Work item created - {rule_id}: {html_url}")
    else:
        print(f"Failed to create work item: {rule_id}")
        sys.exit(1)



#! /usr/bin/env python3

import click
from sarif_manager.bin.logger import init_logger

from sarif_manager.version import __version__
from sarif_manager.command.azure import azure
from dotenv import load_dotenv

load_dotenv()


@click.group
@click.version_option(version=__version__)
def sarif_manager():
    """Parse and use SARIF files for security analysis."""
    init_logger()


sarif_manager.add_command(azure)


def main():
    """Parse and use SARIF files for security analysis."""
    sarif_manager()


if __name__ == '__main__':
    main()



from sarif_manager.sarif.azure_sarif_utils import trim_uuid


class AzureSarifFinding:
    def __init__(self, repo_url: str, rule_id: str, message: str, severity: str, location: str, line: int, description: str):
        self.repo_url = repo_url
        self.original_rule_id = rule_id
        # Now the rule ID will end up being the same as the message.
        self.rule_id = trim_uuid(rule_id)
        self.message = message
        self.severity = 'warning' if severity == 'warning' else 'error'
        self.location = location
        self.line = line
        self.description = description

    @property
    def file_path(self):
        return self.location.replace('\\', '/')

    @property
    def file_url(self):
        return f"{self.repo_url}?path=/{self.file_path}&version=GBmain&line={self.line + 1}&lineEnd={self.line + 2}&lineStartColumn=1&lineEndColumn=1&lineStyle=plain&_a=contents"

    @property
    def formatted_message(self):
        return f"{self.rule_id} at {self.location}:{self.line} | {self.file_url}"

    @property
    def azure_devops_message(self):
        return f"##vso[task.logissue type={self.severity}]{self.formatted_message}"

    @property
    def exclude(self):
        """Exclude certain findings. Missing HTTP Headers, or findings that don't have lines."""
        if "Missing HTTP Header" in self.message:
            return True
        # Exclude findings that don't have a file path
        if self.file_path == "/":
            return True
        # Exclude findings that don't have a line number. It will be set to 0 or 1 if it's missing.
        if self.line == 0 or self.line is None or self.line == 1:
            return True
        return False

    def __str__(self):
        return self.formatted_message



"""Setup script"""
import setuptools
import os
import re

HERE = os.path.abspath(os.path.dirname(__file__))
VERSION_RE = re.compile(r"""__version__ = ['"]([0-9.]+)['"]""")
TESTS_REQUIRE = ["coverage", "pytest"]


def get_version():
    init = open(os.path.join(HERE, "sarif_manager", "version.py")).read()
    return VERSION_RE.search(init).group(1)


def get_description():
    return open(
        os.path.join(os.path.abspath(HERE), "README.md"), encoding="utf-8"
    ).read()


setuptools.setup(
    name="sarif-manager",
    include_package_data=True,
    version=get_version(),
    author="Kinnaird McQuade",
    author_email="kinnaird@nightvision.net",
    description="Parse SARIF files for different providers.",
    packages=setuptools.find_packages(exclude=["test*"]),
    long_description=get_description(),
    long_description_content_type="text/markdown",
    url="https://github.com/nvsecurity/api-validator",
    tests_require=TESTS_REQUIRE,
    install_requires=[
        "beautifulsoup4",
        "click",
        "click-option-group",
        "loguru",
        "markdown",
        "requests",
        "python-dotenv",
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    entry_points={"console_scripts": "sarif-manager=sarif_manager.bin.cli:main"},
    zip_safe=True,
    keywords="sarif",
    python_requires=">=3.11",
)
